module Hikki
  module Adapters
    class RedisCollection < Hikki::Collection
      attr_reader :connection, :uuid_generator

      def initialize(collection, connection, uuid_generator)
        super(collection)
        @connection = connection
        @uuid_generator = uuid_generator
      end

      def index(field)
        connection.sadd collection_indexes_key, field.to_s
        true
      end

      def save(data)
        data = normalize_data(data)
        connection.hset collection_key, data['id'], data.to_json
        add_to_index(data)
        data
      end

      def find(id)
        JSON.parse(connection.hget(collection_key, id.to_s) || '{}')
      end

      def all(options={})
        options = normalize_options(options)
        connection.hvals(collection_key)[page_range(options)].map { |j| JSON.parse(j) }
      end

      def find_by(field, value, options={})
        options = normalize_options(options)
        return find_by_index(field, value, options) if has_index?(field)
        all.select { |o| o.fetch(field.to_s) == value }[page_range(options)]
      end

      def remove(id)
        remove_from_index(find(id))
        connection.hdel collection_key, id.to_s
        true
      end

      def remove_all
        indexes.each { |field| connection.del collection_index_key(field) }
        connection.del collection_indexes_key
        connection.del collection_key
      end

      def has_index?(field)
        connection.sismember collection_indexes_key, field.to_s
      end

      def indexes
        connection.smembers collection_indexes_key
      end

      def id_for(data)
        data.fetch('id', uuid_generator.uuid).to_s
      end

      private
      def collection_indexes_key
        "hikki:#{collection}:indexes"
      end

      def collection_index_key(field)
        "hikki:#{collection}:#{field.to_s}"
      end

      def collection_key
        "hikki:#{collection}"
      end

      def add_to_index(data)
        indexes.each do |field|
          existing = indexed_ids(field, data[field])
          existing << data['id'] unless existing.include? data['id']
          connection.hset collection_index_key(field), data[field], existing.to_json
        end
      end

      def remove_from_index(data)
        return if data == {}
        indexes.each do |field|
          ids = indexed_ids(field, data[field])
          ids.delete(data['id'])
          connection.hset collection_index_key(field), data[field].to_s, ids.to_json
        end
      end

      def indexed_ids(field, value)
        JSON.parse(connection.hget(collection_index_key(field), value.to_s) || '[]')
      end

      def find_by_index(field, value, options)
        ids = indexed_ids(field, value)
        results = connection.pipelined { ids.each { |id| connection.hget collection_key, id } }
        results.map { |j| JSON.parse(j) }.reject { |v| v == {} }[page_range(options)]
      end

    end
  end
end